/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define TArmeMasterMovie_h

#ifndef TArmeMasterMovie_h
#include "TArmeMasterMovie.h"
#endif //TArmeMasterMovie_h

#ifndef ZzArray_h
#include "ZzArray.h"
#endif //ZzArray_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#ifndef ArmeArchivable_h
#include "ArmeArchivable.h"
#endif //ArmeArchivable_h

#ifndef WorkingCodec_h
#include "WorkingCodec.h"
#endif //WorkingCodec_h

class MIProgressBarDlgManager;

#include <Quicktime/Quicktime.h>

#define MAX_TRACK_LEVELS 5
#define MASTER_MOVIE_MAX_DURATION 36000


class TArmeDoc;
class TVideoObject;
class ArmeVideoDataFileRef;
class TArmeMasterMovieOriginalTrack;
class TArmeMasterMovieTrack;
class TArmeMasterMovieChunkHierarchyElement;
class TArmeMasterMovieChunk;
class TArmeMasterMovieHierarchyElement;
class TArmeMasterMovieTextTrack;
class TArmeTextClip;
class ArmeDataFileRef;

enum {
	kVideoMedia,
	kTextMedia,
	kKeyedTextMedia,
};

/*! 
	@class TArmeMasterMovie
	Each document contains a TArmeMasterMovie.  This movie is THE movie in which all the video clips and effects are played.  
	Every video object is a track (or many track) in this movie.
	The frame rate and dimension of this movie shall be set in the project settings and shall not be modified later.
*/
class TArmeMasterMovie : public ArmeArchivable
{
	TArmeDoc*				fDoc;
	
	// file related members
public :
	short					fResourceRefNum;
	short					fResourceId;
	FSSpec 					fsSpec;
protected :

	// codecs related members
	WorkingCodecType		fWorkingCodecType;
	WorkingCodecType		fPreviewCodecType;
	SInt32					fVideoTimeScale;
	SInt32					fVideoFrameDuration;
	
	// QT related members
	Movie					fQTMovie;
	Track					fQTBackgroundTrack;

	// original track that contains a copy of all the movie that have been imported in the workspace
	TArmeMasterMovieOriginalTrack* fOriginalTrack;

	// chunks related members
	ZzArray<ArmeTime>*								fSplitsArray;
	ZzArray<TArmeMasterMovieChunk*>*				fChunksArray;
	ZzArray<TArmeMasterMovieHierarchyElement*>*		fHierarchiesArray;
	
	ZzArray<TArmeMasterMovieTextTrack*>*			fTextTrackArray;

	TArmeMasterMovieTextTrack*	CreateNewTextTrack();
	void					RemoveTextTrack(TArmeMasterMovieTextTrack* textTrack);
	
	void					AddSplit(ArmeTime splitTime);
	void					RemoveSplit(ArmeTime splitTime);

	TArmeMasterMovieChunk*	GetChunk(ArmeTime startTime, ArmeTime endTime);
	TArmeMasterMovieChunk*	GetChunkStartingAtTime(ArmeTime startTime);
	TArmeMasterMovieChunk*	GetChunkAtTime(ArmeTime time);
	TArmeMasterMovieChunk*	GetChunk(int index);
public :
	TArmeMasterMovieTextTrack*	GetTextTrack(int i);

	void							SetPlaybackQuality(bool highQuality, bool singleField);

	void							InsertChunkAfter(TArmeMasterMovieChunk* insertAfter, TArmeMasterMovieChunk* insertObject);
	void							RemoveChunk(TArmeMasterMovieChunk* chunkToRemove);

	void								RemoveHierarchy(TArmeMasterMovieHierarchyElement* hierarchyToRemove);
	void								AddHierarchy(TArmeMasterMovieHierarchyElement* hierarchyToAdd);
	TArmeMasterMovieHierarchyElement* 	GetHierarchyForChunk(TArmeMasterMovieChunk* chunk);
private :
	TArmeMasterMovieHierarchyElement*	CreateNewHierarchyForChunk(TArmeMasterMovieChunkHierarchyElement* chunkElement);
	
	TArmeMasterMovie(TArmeDoc* doc, WorkingCodecType workingCodec, const char* fileName);  // constructor used to build a temp mastermovie
	void Init(TArmeDoc* doc, WorkingCodecType workingCodec, const char* fileName);
	
public :
	SInt32				fTrackIdForTextMedia;

	Track				GetMasterTrackForText();

	Movie				GetQTMovie() const {return fQTMovie;};
	TArmeDoc*			GetDoc() const {return fDoc;};
	int					GetWidth() const;
	int					GetHeight() const;
	int					GetWorkingWidth() const;
	int					GetWorkingHeight() const;
	int					GetViewerWidth() const;
	int					GetViewerHeight() const;
	float				GetFPS() const;
	WorkingCodecType	GetWorkingCodecType() const;
	WorkingCodecType	GetWorkingCodecTypeForPictures() const;
	WorkingCodecType	GetPreviewCodecType() const;
private :
	void				SetWorkingCodecType(WorkingCodecType codec);
	void				SetPreviewCodecType(WorkingCodecType codec);
public :
	TArmeMasterMovieOriginalTrack* 	GetOriginalTrack() const {return fOriginalTrack;};

	SInt32				GetVideoTimeScale() const {return fVideoTimeScale;};
	SInt32				GetVideoFrameDuration() const {return fVideoFrameDuration;};

	void	CreateBackground();

	void	AddSubMovie(TVideoObject* obj);
	void 	RemoveSubMovie(TVideoObject* obj);
	void	UpdateSubMovie(TVideoObject* obj);
	void	RebuildMasterMovie();
	
	void 	UpdateFile() const;
	bool	UpdateForSelfContainedProject(ArmeDataFileRef* dataFileRef, MIProgressBarDlgManager* progressMngr);
	bool	RemoveReferenceTo(ArmeDataFileRef* dataFileRef);
	
	TArmeMasterMovie(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;
	
	TArmeMasterMovie(TArmeDoc* doc, WorkingCodecType workingCodec);
	virtual ~TArmeMasterMovie();
	
	void				CopyBackgroundInTrack(Track destTrack, ArmeTime startTime, ArmeTime endTime);
	
	PicHandle			GetPictAtTime(TVideoObject* obj, TVideoObject* inEffect, ArmeTime time);
};


class TArmeMasterMovieOriginalTrack : public ArmeArchivable
{
	SInt32					fTrackID;
	TimeValue				fCurrentLength;
	ZzArray<TimeValue>*		fFileStartTimeArray;
	ZzArray<TimeValue>*		fFileDurationArray;
	ZzArray<ZzArray<PicHandle>*>*	fFileCachedThumbs;
	ZzArray<ZzArray<TimeValue>*>*	fFileCachedThumbsTime;
	ZzArray<SInt32>*		fIndexArray;
	TArmeMasterMovie*		fMasterMovie;
public :
	
	void 	AddFileInOriginalTrack(ArmeVideoDataFileRef* file, Movie movie, int trackIndex);
	void 	RemoveFileFromOriginalTrack(ArmeVideoDataFileRef* file);
	TArmeMasterMovieOriginalTrack(TArmeMasterMovie* masterMovie);
	virtual ~TArmeMasterMovieOriginalTrack();

	TArmeMasterMovieOriginalTrack(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;
	
	
	TimeValue	GetStartTimeOfFile(const ArmeVideoDataFileRef* file) const;
	TimeValue	GetStartTimeOfNextFile(const ArmeVideoDataFileRef* file) const;

	PicHandle	GetPict(const ArmeVideoDataFileRef* file, ArmeTime localTime);
	Track		GetQTTrack();
protected : 
	SInt32		GetIndex(const ArmeVideoDataFileRef* file) const;
};

class TArmeMasterMovieTrack : public ArmeArchivable
{
	TArmeMasterMovieTrack*	fSourceA;
	TArmeMasterMovieTrack* 	fSourceB;
	UInt32					fTrackType;
	SInt32					fTrackID;
public :
	TArmeMasterMovie*		fMasterMovie;
	
	UInt32 GetType() const {return fTrackType;};
	
	TArmeMasterMovieTrack(TArmeMasterMovie* masterMovie, TArmeMasterMovieTrack*	sourceA, TArmeMasterMovieTrack*	sourceB, int trackLayer, UInt32 trackType);
	virtual ~TArmeMasterMovieTrack();

	TArmeMasterMovieTrack(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;

	TArmeMasterMovieTrack*	GetSourceA() {return fSourceA;};
	TArmeMasterMovieTrack* 	GetSourceB() {return fSourceB;};
	
	Track 	GetQTTrack() const;
};

class TArmeMasterMovieTextTrack : public TArmeMasterMovieTrack
{
public :
	TArmeMasterMovieTextTrack(TArmeMasterMovie* masterMovie);
	virtual ~TArmeMasterMovieTextTrack();

	TArmeMasterMovieTextTrack(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;
	
	void							AddTextClip(TArmeTextClip* textClip, ArmeTime startTime, ArmeTime endTime);
	void							Clean(ArmeTime startTime, ArmeTime endTime);
};


class TArmeMasterMovieChunk : public ArmeArchivable
{

public :
	TArmeMasterMovieChunkHierarchyElement*	fTopChunk;
	TArmeMasterMovie* 						fMasterMovie;
	ArmeTime								fStartTime;
	ArmeTime								fEndTime;
	TArmeMasterMovieHierarchyElement*		fHierarchy;
	ZzArray<TArmeTextClip*>*				fTextClipsArray;
	ZzArray<TVideoObject*>*					fTextClipOwnerArray;
	
	TArmeMasterMovieHierarchyElement*		GetHierarchy() const {return fHierarchy;};
	
	void	Split(ArmeTime splitTime);
	bool	MergeWith(TArmeMasterMovieChunk* mergeWith);
	
	TArmeMasterMovieChunkHierarchyElement*	FindChunkHierarchyElementHosting(TVideoObject* object);
	void	InsertPieceOfVideoObject(TVideoObject* object, TVideoObject* parentObject);
	
	void	AddTextClip(TArmeTextClip* textClip);
	void	AddRefOnTextClipOwner(TVideoObject* container);
	void	RemoveAllTextClips();
	bool	CompareTextClips(TArmeMasterMovieChunk* chunk);
	
	TArmeMasterMovieChunk(TArmeMasterMovie* masterMovie, ArmeTime startTime, ArmeTime endTime);
	virtual ~TArmeMasterMovieChunk();

	TArmeMasterMovieChunk(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;
};


class TArmeMasterMovieChunkHierarchyElement : public ArmeArchivable
{

public :
	TArmeMasterMovieChunk* 					fMasterMovieChunk;
	TVideoObject*							fVideoObject;
	TVideoObject*							fParentObject;
	TArmeMasterMovieChunkHierarchyElement*	fSourceA;
	TArmeMasterMovieChunkHierarchyElement*	fSourceB;
	TArmeMasterMovieChunkHierarchyElement*	fParent;
	
	UInt32	GetType() const;

	//constructors
	//************
	TArmeMasterMovieChunkHierarchyElement(TVideoObject* videoObject, TVideoObject* parentObject, TArmeMasterMovieChunk* masterMovieChunk, TArmeMasterMovieChunkHierarchyElement* parent);
	virtual ~TArmeMasterMovieChunkHierarchyElement();

	//archiving
	//*********
	TArmeMasterMovieChunkHierarchyElement(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;
	
	TArmeMasterMovieChunkHierarchyElement*	GetCopyOfHierarchy(TArmeMasterMovieChunk* inMasterMovieChunk) const;
	static bool CompareHierarchy(const TArmeMasterMovieChunkHierarchyElement* arg1, const TArmeMasterMovieChunkHierarchyElement* arg2);
	
	TArmeMasterMovieChunkHierarchyElement*	FindChunkHierarchyElementHosting(TVideoObject* object);
	void									InsertChildrenOf(TVideoObject* object);//, TVideoObject* parentObject);

	TArmeMasterMovieChunkHierarchyElement* CreateTransitions(ZzArray<TVideoObject*>* sortedChildren, TVideoObject* parentObject, TArmeMasterMovieChunk* masterMovieChunk);
};

class TArmeMasterMovieHierarchyElement : public ArmeArchivable
{
	TArmeMasterMovieHierarchyElement*		fSourceA;
	TArmeMasterMovieHierarchyElement*		fSourceB;
	TArmeMasterMovieTrack*					fMasterMovieTrack;
	SInt32									fRefCount;


public :

	void 		SetPlaybackQuality(bool highQuality, bool singleField);


	UInt32	GetType() const {return fMasterMovieTrack->GetType();};
	void	AddRef() {fRefCount++;};
	SInt32	Release();

	void	Clean(ArmeTime startTime, ArmeTime endTime);
	void	CopyChunk(TArmeMasterMovieChunkHierarchyElement* chunkElement);

	bool	ContainsTextClips() const;
	bool	CanHostChunk(TArmeMasterMovieChunkHierarchyElement* chunk);
	TArmeMasterMovieTrack*					GetMasterMovieTrack() const {return fMasterMovieTrack;};	
	
	TArmeMasterMovieHierarchyElement(TArmeMasterMovie* masterMovie, TArmeMasterMovieHierarchyElement* sourceA, TArmeMasterMovieHierarchyElement* sourceB, UInt32 trackType);
	virtual ~TArmeMasterMovieHierarchyElement();
	
	//archiving
	//*********
	TArmeMasterMovieHierarchyElement(ArmeArchiveHandler* unarchiver);
	virtual bool WriteToArchive(ArmeArchiveHandler* archiver) const;
};